package sygr3emif;

import java.util.ArrayList;

import sygr.pots.extensions.Attr;
import sygr.pots.extensions.NodeValue;
import sygr.pots.extensions.PluginData;
import sygr.pots.extensions.PluginInterface;
import sygr.pots.extensions.PluginUtilInterface;
import sygr.pots.extensions.Pot;
import sygr3em.model.SingleTransfer;
import sygr3em.service.S3eConstants;

public class SyncOutTestMd implements PluginInterface {
	
	private String LOGPREFIX = "S3eDemo: ";

	@Override
	public void execute(PluginData data, PluginUtilInterface util) {
		
		if(data == null || util == null) {
			util.log(LOGPREFIX + "data or util is null.");
			return;
		}
		if(data.transfer == null || data.transfer.size() == 0
				|| data.response == null) {
			util.log(LOGPREFIX + "No content to work with.");
			return;
		}
		
		// Collect what we have
		SingleTransfer httpstatus = null;
		SingleTransfer id = null;
		ArrayList<SingleTransfer> errors = new ArrayList<>();
		for(Object ob: data.transfer) {
			if(ob instanceof SingleTransfer) {
				SingleTransfer st = ( SingleTransfer ) ob;
				if(st.getType().equals(S3eConstants.HTTPSTATUS)) {
					httpstatus = st;
				}
				if(st.getType().equals(S3eConstants.ID)) {
					id = st;
				}
				if(st.getType().equals(S3eConstants.DATA)) {
					errors.add(st);
				}
			}
		}
		
		// Have ID?
		if(id == null) {
			util.log(LOGPREFIX + "No ID information arrived");
			data.response.setStatus(406);
			return;
		}
		
		// Set the status
		int st = 406;
		if(!(httpstatus == null) && httpstatus.getType().equals(S3eConstants.HTTPSTATUS)) {
			try {
				st = Integer.parseInt(httpstatus.getValue());
			} catch(Exception e) {}
			data.response.setStatus(st);
		}
				
		// Get the text lines. If nothing, then the master data is fine.
		if(errors == null || errors.size() == 0) {
			data.response.setAnswer("EXTERNAL TEST: Master data " + id.getValue() 
			+ " for environmental variable " + id.getKey() + " is perfect.");
			return;
		}
		
		String ans = "EXTERNAL TEST found the following errors:"
				 + System.lineSeparator() + System.lineSeparator();
		
		for(SingleTransfer error: errors) {
			ans = ans + error.getKey() + ": " + error.getValue() + System.lineSeparator();
		}
		data.response.setAnswer(ans);
		
	}

	@Override
	public void reloadConfig(ArrayList<String> changes, PluginUtilInterface util) {}

}
